#include "ttl_p.h"

enum {
  ARY_EXTEND = 50,
};

// static
int        TtWindow::wnd_count    = 0;
int        TtWindow::ary_max      = 0;
HWND      *TtWindow::h_wnd_ary    = NULL;
TtWindow **TtWindow::instance_ary = NULL;

LRESULT CALLBACK TtWindow::_ttl_wnd_proc( HWND h_wnd, UINT msg, WPARAM w_param, LPARAM l_param ) {
  for ( int i = 0; i < wnd_count; ++i ) {
    if ( h_wnd_ary[i] == h_wnd ) {
      return instance_ary[i]->wnd_proc( msg, w_param, l_param );
    }
  }
  return DefWindowProc( h_wnd, msg, w_param, l_param );
}

void TtWindow::init_common_controls() {
  InitCommonControls();
}

// override at sub class
DWORD  TtWindow::get_ex_style()   { return 0; }
DWORD  TtWindow::get_style()      { return 0; }
PCTSTR TtWindow::get_class_name() { return NULL; }
bool TtWindow::created() { return true; }

LRESULT TtWindow::wnd_proc( UINT msg, WPARAM w_param, LPARAM l_param ) {
  if ( this->wnd_proc_super == NULL ) {
    return DefWindowProc( this->h_wnd, msg, w_param, l_param );
  }
  else {
    return CallWindowProc( wnd_proc_super, this->h_wnd, msg, w_param, l_param );
  }
}


// instance
TtWindow::TtWindow() {
  this->h_wnd = NULL;
  this->id = 0;
  this->h_instance = NULL;
  this->wnd_proc_super = NULL;
}

bool TtWindow::create( TtWindow *parent, int id, HINSTANCE h_instance, LPVOID lp_param ) {
  this->parent = parent;
  this->h_instance = h_instance;
  this->id = id;
  
  this->h_wnd = CreateWindowEx(
    this->get_ex_style(),
    this->get_class_name(),
    NULL,
    ( parent == NULL ) ? (this->get_style()) : (this->get_style() | WS_CHILD),
    CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
    parent ? parent->get_h_wnd() : NULL,
    (HMENU)id,
    h_instance,
    lp_param );
  if ( this->h_wnd == NULL ) { return false; }

  ++TtWindow::wnd_count;
  if ( TtWindow::wnd_count > TtWindow::ary_max ) {
    int new_max = TtWindow::ary_max + ARY_EXTEND;
    HWND *new_wnd = new HWND[new_max];
    TtWindow **new_ins = new TtWindow*[new_max];
    
    for ( int i = 0; i < TtWindow::ary_max; ++i ) {
      new_wnd[i] = TtWindow::h_wnd_ary[i];
      new_ins[i] = TtWindow::instance_ary[i];
    }

    TtWindow::ary_max = new_max;
    delete [] TtWindow::h_wnd_ary;
    TtWindow::h_wnd_ary = new_wnd;
    delete [] TtWindow::instance_ary;
    TtWindow::instance_ary = new_ins;
  }
  TtWindow::h_wnd_ary[TtWindow::wnd_count - 1]    = this->h_wnd;
  TtWindow::instance_ary[TtWindow::wnd_count - 1] = this;

  if ( !(this->created()) ) { return false; }

  return true;
}

int TtWindow::close() {
  return SendMessage( this->h_wnd, WM_CLOSE, 0, 0 );
}

HWND      TtWindow::get_h_wnd()      { return this->h_wnd; }
int       TtWindow::get_id()         { return this->id; }
HINSTANCE TtWindow::get_h_instance() { return this->h_instance; }

TtWindow *TtWindow::get_parent() {
  return this->parent;
}


RECT TtWindow::get_rect() {
  RECT rect;
  GetWindowRect( this->h_wnd, &rect );
  return rect;
}

int TtWindow::get_pos_x() { return this->get_rect().left; }
int TtWindow::get_pos_y() { return this->get_rect().top; }

int TtWindow::get_width() {
  RECT rect = this->get_rect();
  return rect.right - rect.left;
}
int TtWindow::get_height() {
  RECT rect = this->get_rect();
  return rect.bottom - rect.top;
}


int TtWindow::show() {
  return ShowWindow( this->h_wnd, SW_SHOW );
}

int TtWindow::hide() {
  return ShowWindow( this->h_wnd, SW_HIDE );
}

int TtWindow::minimize() {
  return ShowWindow( this->h_wnd, SW_MINIMIZE );
}

int TtWindow::maximize() {
  return ShowWindow( this->h_wnd, SW_MAXIMIZE );
}

int TtWindow::restore() {
  return ShowWindow( this->h_wnd, SW_RESTORE );
}

int TtWindow::set_show_state( int state ) {
  return ShowWindow( this->h_wnd, state );
}


int TtWindow::set_position( int x, int y ) {
  RECT rect;
  GetWindowRect( this->h_wnd, &rect );
  return MoveWindow( this->h_wnd, x, y, rect.right - rect.left, rect.bottom - rect.top, TRUE );
}

int TtWindow::set_size( int width, int height ) {
  RECT rect;
  GetWindowRect( this->h_wnd, &rect );
  return MoveWindow( this->h_wnd, rect.left, rect.top, width, height, TRUE );
}

int TtWindow::set_pos_size( int x, int y, int width, int height ) {
  return MoveWindow( this->h_wnd, x, y, width, height, TRUE );
}

int TtWindow::set_rect( RECT rect ) {
  return MoveWindow( this->h_wnd, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, TRUE );
}


bool TtWindow::set_bottom() {
  return SetWindowPos( this->h_wnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE );
}

bool TtWindow::set_top() {
  return SetWindowPos( this->h_wnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE );
}

bool TtWindow::set_under( HWND wnd ) {
  return SetWindowPos( this->h_wnd, wnd, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE );
}


int TtWindow::set_font( HFONT h_font ) {
  return SendMessage( this->h_wnd, WM_SETFONT, (WPARAM)h_font, (LPARAM)true );
}


int TtWindow::set_forground() {
  return SetForegroundWindow( this->h_wnd );
}


int TtWindow::set_text( const char *str ) {
  return SetWindowText( this->h_wnd, str );
}

int TtWindow::get_text( char *buf, int buf_size ) {
  return GetWindowText( this->h_wnd, buf, buf_size );
}

int TtWindow::messageloop() {
  MSG msg;
  while ( GetMessage( &msg, NULL, 0, 0 ) ) {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }
  return msg.wParam;
}

int TtWindow::override_wnd_proc() {
  this->wnd_proc_super = (WNDPROC)GetWindowLong( this->h_wnd, GWL_WNDPROC );
  return SetWindowLong( this->h_wnd, GWL_WNDPROC, (LONG)( TtWindow::_ttl_wnd_proc ) );
}
